BEGIN;

DROP TABLE IF EXISTS graph_version CASCADE;
DROP TABLE IF EXISTS graph_node CASCADE;
DROP TABLE IF EXISTS graph_link CASCADE;
DROP TABLE IF EXISTS graph_iterator_result CASCADE;

CREATE TABLE graph_version (version VARCHAR(32));
INSERT INTO graph_version VALUES ('1.01');

CREATE TABLE graph_node (
	node_id VARCHAR(1024) PRIMARY KEY,
	node_type VARCHAR(1024) NOT NULL,
	node_group VARCHAR(1024),
	node_label TEXT,
	node_seq BIGSERIAL
);

CREATE INDEX graph_node_group ON graph_node (node_group);
CREATE INDEX graph_node_seq ON graph_node (node_seq);

CREATE TABLE graph_link (
	link_src_id VARCHAR(1024) NOT NULL REFERENCES graph_node (node_id),
	link_dst_id VARCHAR(1024) NOT NULL REFERENCES graph_node (node_id),
	link_type VARCHAR(1024) NOT NULL,
	link_src_group VARCHAR(1024),
	link_dst_group VARCHAR(1024),
	link_weight DOUBLE PRECISION NOT NULL DEFAULT 1.0,
	link_label TEXT,
	PRIMARY KEY (link_src_id, link_dst_id, link_type)
);

CREATE INDEX graph_link_link_src_id_link_type ON graph_link (link_src_id, link_type);
CREATE INDEX graph_link_link_dst_id_link_type ON graph_link (link_dst_id, link_type);
CREATE INDEX graph_link_link_src_group_link_type ON graph_link (link_src_group, link_type);
CREATE INDEX graph_link_link_dst_group_link_type ON graph_link (link_dst_group, link_type);


CREATE TABLE graph_iterator_result (
	id VARCHAR(1024),
	is_group BOOLEAN,
	seq BIGINT
);



CREATE OR REPLACE FUNCTION add_node(id VARCHAR(1024), ntype VARCHAR(1024), label TEXT) RETURNS BIGINT AS $$
	DECLARE
		cur_type VARCHAR(1024);
	BEGIN
		SELECT node_type INTO cur_type FROM graph_node WHERE node_id = $1;
		IF cur_type IS NULL THEN
			INSERT INTO graph_node (node_id, node_type, node_label) VALUES ($1, $2, $3);
		ELSIF cur_type = $2 THEN
			UPDATE graph_node SET node_label = $3 WHERE node_id = $1;
		ELSE
			DELETE FROM graph_link WHERE link_src_id = $1 OR link_dst_id = $1;
			DELETE FROM graph_node WHERE node_id = $1;
			INSERT INTO graph_node (node_id, node_type, node_label) VALUES ($1, $2, $3);
		END IF;
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION add_link(src_id VARCHAR(1024), dst_id VARCHAR(1024), ltype VARCHAR(1024), weight DOUBLE PRECISION, label TEXT) RETURNS BIGINT AS $$
	DECLARE
		link_exists BOOLEAN;
		src_group VARCHAR(1024);
		dst_group VARCHAR(1024);
	BEGIN
		SELECT COUNT(*) > 0 INTO link_exists FROM graph_link WHERE link_src_id = $1 AND link_dst_id = $2 AND link_type = $3;
		IF link_exists THEN
			UPDATE graph_link SET link_weight = $4, link_label = $5 WHERE link_src_id = $1 AND link_dst_id = $2 AND link_type = $3;
		ELSE
			SELECT node_group INTO src_group FROM graph_node WHERE node_id = $1;
			SELECT node_group INTO dst_group FROM graph_node WHERE node_id = $2;
			INSERT INTO graph_link (link_src_id, link_dst_id, link_type, link_src_group, link_dst_group, link_weight, link_label) VALUES ($1, $2, $3, src_group, dst_group, $4, $5);
		END IF;
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION set_group(id VARCHAR(1024), group_id VARCHAR(1024)) RETURNS BIGINT AS $$
	BEGIN
		UPDATE graph_node SET node_group = $2 WHERE node_id = $1;
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION finish_group(group_id VARCHAR(1024)) RETURNS BIGINT AS $$
	DECLARE
		types INTEGER;
	BEGIN
		SELECT COUNT(*) INTO types FROM (SELECT DISTINCT node_type FROM graph_node WHERE node_group = $1 GROUP BY node_type) AS foo;
		IF types <> 1 THEN
			UPDATE graph_node SET node_group = NULL WHERE node_group = $1;
			RETURN 1;
		END IF;
		
		UPDATE graph_link SET link_src_group = $1 WHERE link_src_id IN (SELECT node_id FROM graph_node WHERE node_group = $1);
		UPDATE graph_link SET link_dst_group = $1 WHERE link_dst_id IN (SELECT node_id FROM graph_node WHERE node_group = $1);
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION remove_node(id VARCHAR(1024)) RETURNS BIGINT AS $$
	BEGIN
		DELETE FROM graph_link WHERE link_src_id = $1 OR link_dst_id = $1;
		DELETE FROM graph_node WHERE node_id = $1;
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION remove_link(src_id VARCHAR(1024), dst_id VARCHAR(1024), ltype VARCHAR(1024)) RETURNS BIGINT AS $$
	BEGIN
		DELETE FROM graph_link WHERE link_src_id = $1 AND link_dst_id = $2 AND link_type = $3;
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION clear_graph() RETURNS BIGINT AS $$
	BEGIN
		DELETE FROM graph_link;
		DELETE FROM graph_node;
		
		RETURN 0;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION get_atomic_nodes_by_node_id(id VARCHAR(1024)) RETURNS SETOF graph_node AS $$
	DECLARE
		row graph_node%ROWTYPE;
	BEGIN
		SELECT * INTO row FROM graph_node WHERE node_id = $1;
		IF row IS NULL THEN
			RETURN;
		END IF;
		
		IF row.node_group IS NULL THEN
			RETURN QUERY SELECT row.node_id, row.node_type, row.node_group, row.node_label, row.node_seq;
		ELSE
			RETURN QUERY SELECT * FROM graph_node WHERE node_group = row.node_group;
		END IF;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION get_atomic_nodes_by_node_group(group_id VARCHAR(1024)) RETURNS SETOF graph_node AS $$
	DECLARE
		row graph_node%ROWTYPE;
	BEGIN
		RETURN QUERY SELECT * FROM graph_node WHERE node_group = $1;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION get_incident_links_by_node_id(id VARCHAR(1024)) RETURNS SETOF graph_link AS $$
	BEGIN
		RETURN QUERY SELECT * FROM graph_link WHERE link_src_id = $1 OR link_dst_id = $1;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION get_incident_links_by_node_group(group_id VARCHAR(1024)) RETURNS SETOF graph_link AS $$
	BEGIN
		RETURN QUERY SELECT * FROM graph_link WHERE link_src_group = $1 OR link_dst_group = $1;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION get_all_nodes(min_seq BIGINT, batch INTEGER) RETURNS SETOF graph_iterator_result AS $$
	BEGIN
		RETURN QUERY SELECT * FROM (
				(SELECT node_id AS id, FALSE AS is_group, node_seq AS seq FROM graph_node WHERE node_group IS NULL AND node_seq >= min_seq ORDER BY seq LIMIT batch)
			UNION
				(SELECT node_group AS id, TRUE AS is_group, MAX(node_seq) AS seq FROM graph_node WHERE node_group IS NOT NULL AND node_seq >= min_seq GROUP BY node_group ORDER BY seq LIMIT batch)
			) AS foo ORDER BY seq LIMIT batch;
	END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION get_all_nodes(ntype VARCHAR(1024), min_seq BIGINT, batch INTEGER) RETURNS SETOF graph_iterator_result AS $$
	BEGIN
		RETURN QUERY SELECT * FROM (
				(SELECT node_id AS id, FALSE AS is_group, node_seq AS seq FROM graph_node WHERE node_type = ntype AND node_group IS NULL AND node_seq >= min_seq ORDER BY seq LIMIT batch)
			UNION
				(SELECT node_group AS id, TRUE AS is_group, MAX(node_seq) AS seq FROM graph_node WHERE node_type = ntype AND node_group IS NOT NULL AND node_seq >= min_seq GROUP BY node_group ORDER BY seq LIMIT batch)
			) AS foo ORDER BY seq LIMIT batch;
	END;
$$ LANGUAGE plpgsql;

COMMIT;